'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _mapValues = _interopDefault(require('lodash/mapValues'));
var _omit = _interopDefault(require('lodash/omit'));
var _sumBy = _interopDefault(require('lodash/sumBy'));
var _toString = _interopDefault(require('lodash/toString'));
var _map = _interopDefault(require('lodash/map'));
var dateFns = require('date-fns');
var _defaults = _interopDefault(require('lodash/defaults'));
require('isomorphic-fetch');
var Path = _interopDefault(require('path'));
var Url = _interopDefault(require('url'));
var fse = _interopDefault(require('fs-extra'));
var _toPairs = _interopDefault(require('lodash/toPairs'));
var _sortBy = _interopDefault(require('lodash/sortBy'));
var PdfmakePrinter = _interopDefault(require('pdfmake'));
var Stream = _interopDefault(require('stream'));

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

// copy from https://github.com/vkiryukhin/jsonfn
const toJSON = value => {
  if (value instanceof Function || typeof value === 'function') {
    const fnBody = value.toString();

    if (fnBody.length < 8 || fnBody.substring(0, 8) !== 'function') {
      // this is ES6 Arrow Function
      return `_EsFunc_${fnBody}`;
    }

    return fnBody;
  }

  return value;
};

const fromJSON = value => {
  if (typeof value === 'string' && value.length >= 8) {
    const prefix = value.substring(0, 8);

    if (prefix === 'function') {
      // eslint-disable-next-line no-eval
      return eval(`(${value})`);
    }

    if (prefix === '_EsFunc_') {
      // eslint-disable-next-line no-eval
      return eval(value.slice(8));
    }
  }

  return value;
};

const parseObject = value => {
  if (value) {
    const type = typeof value;

    if (type === 'string') {
      return fromJSON(value);
    }

    if (type === 'object' && !Array.isArray(value)) {
      return _mapValues(value, parseObject);
    }
  }

  return value;
};

var JSONfn = {
  toJSON,
  fromJSON,
  parseObject,
  stringify: obj => JSON.stringify(obj, (key, value) => toJSON(value)),
  parse: str => JSON.parse(str, (key, value) => fromJSON(value))
};

const util = {
  map: _map,
  toString: _toString,
  sumBy: _sumBy,
  dateFnsFormat: dateFns.format // const specialKeys = ['render', 'layout', 'baseKey', '_base']

};
const render = (data, template) => {
  const templateRender = template.render;

  if (templateRender) {
    if (typeof templateRender === 'function') {
      return templateRender.call(util, data, _omit(template, 'render'), util);
    }

    return templateRender;
  } // render list of sub-template which ordered by layout


  if (template.layout) {
    return _map(template.layout, k => render(data, _objectSpread({}, template, template[k])));
  }

  return template;
};
util.render = render;

const _extendDeep = (a, b) => {
  if (b && typeof a === 'object' && !Array.isArray(a)) {
    return _objectSpread({}, b, {
      _base: b
    }, _mapValues(a, (subA, key) => {
      const subB = b[subA.baseKey || key];
      return subB ? _extendDeep(subA, subB) : subA;
    }));
  }

  return a;
};

const extend = (a, b) => {
  const template = _extendDeep(a, b); // handle main block


  const main = template.main || template.Main;

  if (main) {
    return _objectSpread({}, _omit(template, 'main', 'Main'), main);
  }

  return template;
};
util.extend = extend;

async function fetchJson(url, option = {}) {
  const headers = option.headers = _defaults(option.headers, {
    Accept: 'application/json',
    'Content-Type': 'application/json'
  });

  if (option.body && headers['Content-Type'] === 'application/json') {
    option.body = JSON.stringify(option.body);
  }

  return Promise.race([fetch(url, option).then(res => {
    const rType = res.headers.get('Content-Type');
    const promise = !rType || rType.indexOf('/json') >= 0 ? res.text() : Promise.resolve(res);
    if (res.status >= 200 && res.status < 400) return promise; // console.log('>>>', url, res.statusText)

    return promise.then(data => {
      const error = new Error(data && data.message || res.statusText);
      Object.assign(error, {
        response: res,
        data
      });
      return Promise.reject(error);
    });
  }), new Promise((x, reject) => setTimeout(() => reject(new Error('Timeout')), option.timeout || 3000))]);
}
const tryCache = (cache, key, func) => {
  const c = cache[key];
  if (c) return c;
  return cache[key] = func(); // eslint-disable-line
};

const cacheDir = Path.join(__dirname, '../.cache');
let memoryPromises = {};

const outputStream = (cacheFilepath, stream) => new Promise(resolve => {
  fse.ensureDir(cacheDir).then(() => {
    stream.pipe(fse.createWriteStream(cacheFilepath)).on('finish', resolve);
  });
});

const serverFetch = (url, option, postProcessor) => tryCache(memoryPromises, url, () => {
  const urlObj = new Url.URL(url);
  const cacheFilepath = Path.join(cacheDir, Path.basename(urlObj.pathname)); // api json return should have no extension, font file should have extension

  const ext = Path.extname(urlObj.pathname);
  const isJsonFile = !ext || ext === '.json'; // console.log('> serverFetch', url, cacheFilepath)

  let memoryP = fetchJson(url, option).then(async res => {
    if (isJsonFile) {
      // console.log('> serverFetch then', isJsonFile, res)
      // json template
      await fse.outputFile(cacheFilepath, res);
      return JSONfn.parse(res);
    } // stream font


    await outputStream(cacheFilepath, res.body);
    return true;
  }).catch(err => {
    if (isJsonFile) {
      // json template
      return fse.readFile(cacheFilepath, 'utf8').then(cache => cache ? JSONfn.parse(cache) : Promise.reject(err));
    } // stream font


    return fse.exists(cacheFilepath).then(cache => cache || Promise.reject(err));
  });

  if (postProcessor) {
    memoryP = memoryP.then(postProcessor);
  }

  return memoryP;
});

const HOST = 'https://ibill.today';

//   console.log(obj)
//   return obj
// }

const _printers = {};

const getPrinter = prop => {
  const {
    fonts
  } = prop.template.extended;

  const fontArr = _sortBy(_toPairs(fonts), 0);

  const fontKey = JSON.stringify(fontArr);
  return tryCache(_printers, fontKey, () => {
    const fontNames = fontArr.map(fontPair => fontPair[1]);
    const promises = fontNames.map(fontName => prop.loadFont(_objectSpread({}, prop, {
      fontName
    })));
    return Promise.all(promises).then(() => new PdfmakePrinter({
      Roboto: {
        normal: `${prop.fontDir}/${fonts.normal}`,
        bold: `${prop.fontDir}/${fonts.bold}`,
        italics: `${prop.fontDir}/${fonts.italics}`,
        bolditalics: `${prop.fontDir}/${fonts.boldItalics}`
      }
    }));
  });
};

const makePdf = prop => getPrinter(prop).then(printer => {
  const pdfKitDocument = printer.createPdfKitDocument(prop.pdfDefinition);
  const pdfStream = pdfKitDocument.pipe(Stream.PassThrough());
  pdfKitDocument.end();
  pdfStream.pdfKitDocument = pdfKitDocument;
  return pdfStream;
}); // required: data, fontDir
// load from web: templateId, auth
// load from db: loadTemplate, loadFont,

const _renderPdf = prop => prop.loadTemplate(prop).then(template => {
  const pdfDefinition = render(prop.data, template.extended);
  return makePdf(_objectSpread({}, prop, {
    pdfDefinition,
    template
  }));
});

const loadTemplate = prop => serverFetch(`${prop.HOST}/api/template/${prop.template || prop.templateId}?auth=${prop.auth}`, prop);
const loadFont = prop => serverFetch(`${prop.HOST}/font/${prop.fontName}?auth=${prop.auth}`, prop);
const defaultProp = {
  loadTemplate,
  loadFont,
  HOST,
  fontDir: cacheDir
};
const renderPdf = _prop => _renderPdf(_objectSpread({}, defaultProp, _prop));

exports.render = render;
exports.extend = extend;
exports.cacheFetch = serverFetch;
exports.Json = JSONfn;
exports.makePdf = makePdf;
exports.loadTemplate = loadTemplate;
exports.loadFont = loadFont;
exports.renderPdf = renderPdf;
